# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is mozilla.org code.
#
# Contributor(s):
#   Chris Jones <jones.chris.g@gmail.com>
#
# Alternatively, the contents of this file may be used under the terms of
# either of the GNU General Public License Version 2 or later (the "GPL"),
# or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import string, StringIO

MAKEFILE_TEMPLATE = string.Template(open('Makefile.gen.template.in', 'r').read())
REWRITER_TEMPLATE = string.Template(open('porky.gen.template.cc', 'r').read())

_verbosity = 1
def log(minv, fmt, *args):
    if _verbosity >= minv:
        print >>sys.stderr, fmt % args

class Rewriter:
    template = REWRITER_TEMPLATE

    def __init__(self, name, typerws, callrws):
        self.name = name
        self.typerws = typerws
        self.callrws = callrws
        self._uid = 0

    def uid(self):
        tuid = self._uid
        self._uid += 1
        return tuid

    def gencode(self, out):
        out.write(self.template.substitute({
            'pythonrepr': repr(self),
            'rewriter': self.name,
            'type_rule_decls': self.genTypeRuleDecls(),
            'call_rule_decls': self.genCallRuleDecls(),
        }))

    def genTypeRuleDecls(self):
        s = '\n'.join(
            [ trw.gendecls(self.uid()) for trw in self.typerws ])
        s += ('\nTypeRule _type_rules[] = { '
              + ',\n'.join([ trw.genrule() for trw in self.typerws ]) 
              + '%s%s };'% (',\n' if len(self.typerws) > 0 else '',
                            TypeRewrite.gennullrule()))
        return s

    def genCallRuleDecls(self):
        s = '\n'.join(
            [ crw.gendecls(self.uid()) for crw in self.callrws ])
        s += ('\nCallRule _call_rules[] = {\n'
              + ',\n'.join([ crw.genrule() for crw in self.callrws ])
              + '%s%s };'% (',\n' if len(self.callrws) > 0 else '',
                            CallRewrite.gennullrule()))
        return s

    def __repr__(self):
        s = '<Rewriter name='+ repr(self.name) +'\n'
        for typerw in self.typerws:
            s += '  '+ repr(typerw) +'\n'
        s += '\n'
        for callrw in self.callrws:
            s += '  '+ repr(callrw) +'\n'
        return s +'>'

    @classmethod
    def create(cls, name, rewrites):
        # shuffle rewrites by type, and sort them by decreasing specificity
        typerws = [ rw for rw in rewrites if isinstance(rw, TypeRewrite) ]
        typerws.sort(cmp=lambda x, y: cmp(y, x))
        callrws = [ rw for rw in rewrites if isinstance(rw, CallRewrite) ]
        callrws.sort(cmp=lambda x, y: cmp(y, x))
        return Rewriter(name, typerws, callrws)


class CxxQualifiedId:
    def __init__(self, id):
        self.id = id
        self.scopes = [ ]

    def qualify(self, id):
        self.scopes.append(self.id)
        self.id = id

    def iterparts(self):
        for scope in self.scopes:
            yield scope
        yield self.id

    def numparts(self):
        return len(self.scopes) + 1

    def __getitem__(self, i):
        np = self.numparts()
        if i < (np-1): return self.scopes[i]
        elif i == (np-1):    return self.id
        else:            raise IndexError

    def __cmp__(self, o):
        '''"<" means less specific'''
        npself, npo = self.numparts(), o.numparts()
        for i in xrange(min(npself, npo)):
            if self[i] != o[i]:
                return cmp(self[i], o[i]) if npself == npo else npself - npo
        return npself - npo
    def __hash__(self):
        return hash(repr(self))
    def __repr__(self):
        return reduce(
            lambda a, s: repr(s) if not a else a +'::'+ repr(s), 
            self.iterparts(), '')
    def __str__(self):
        return reduce(
            lambda a, s: str(s) if not a else a +'::'+ str(s), 
            self.iterparts(), '')


class CxxType:
    POINTER = ('*',)
    REFERENCE = ('&',)
    ARRAY = ('[]',)

    def __init__(self, base_name):
        self.base_name = base_name # CxxQualifiedId
        self.qualifiers = [ ]
        self.terminal = False   # no more qualifiers allowed

    def pointer(self):
        assert not self.terminal
        self.qualifiers.append(self.POINTER)

    def reference(self):
        assert not self.terminal
        self.qualifiers.append(self.REFERENCE)
        self.terminal = True

    def array(self):
        assert not self.terminal
        self.qualifiers.append(self.ARRAY)

    def __cmp__(self, o): 
        '''"<" means less specific'''
        selfqlen, oqlen = len(self.qualifiers), len(o.qualifiers)
        if selfqlen != oqlen:
            return selfqlen - oqlen
        else:
            return cmp(self.base_name, o.base_name)
    def __hash__(self):  
        return hash(repr(self))
    def __repr__(self):
        r = repr(self.base_name)
        for q in self.qualifiers:  r += repr(q[0])
        return r
    def __repr__(self):
        r = str(self.base_name)
        for q in self.qualifiers:  r += str(q[0])
        return r


class CxxExprSelect:
    DOT = ('.',)
    PDOT = ('->',)

    def __init__(self, base, seltype, field):
        self.base = base
        self.seltype = seltype
        self.field = field

    def __cmp__(self, o):
        return cmp(self.base, o.base) or cmp(self.field, o.field) or cmp(self.seltype, o.seltype)
    def __hash__(self):
        return hash(repr(self))
    def __repr__(self):
        return '%r%s%r'% (self.base, self.seltype[0], self.field)


class CxxExprCall:
    def __init__(self, callable, args, 
                 ismethodcall=False, isnew=False, isdelete=False):
        self.callable = callable
        self.args = args
        self.ismethodcall = ismethodcall # so we can re-sugar if necessary
        self.isnew = isnew
        self.isdelete = isdelete

    def __cmp__(self, o):
        '''"<" means less specific'''
        if (o.isnew and not self.isnew
            or o.isdelete and not self.isdelete):
            return -1
        elif (self.isnew and not o.isnew
            or self.isdelete and not o.isdelete):
            return 1
        elif self.callable < o.callable:
            return -1
        elif o.callable > self.callable:
            return 1
        else:
            if (self.args is None) and (o.args is not None):
                return -1
            elif o.args is None:
                return 1
            elif len(self.args) != len(o.args):
                # not comparable, but makes the sort look nicer
                return len(self.args) - len(o.args)
            else:
                npatsself = reduce(
                    lambda a, p: a + (1 if (isinstance(p, Pattern) 
                                            and p.iswildcard()) else 0),
                    self.args, 0)
                npatso = reduce(
                    lambda a, p: a + (1 if (isinstance(p, Pattern) 
                                            and p.iswildcard()) else 0),
                    o.args, 0)
                return npatsself - npatso
    def __hash__(self):
        return hash(repr(self))
    def __repr__(self):
        if self.isdelete:
            return 'delete %r'% (self.args[0])
        return '%s%r(%r)'% (
            'new ' if self.isnew else '', self.callable, self.args)
    def __str__(self):
        if self.isdelete:
            return 'delete %r'% (self.args[0])
        return '%s%s(%s)'% (
            'new ' if self.isnew else '', self.callable, self.args)


class TypeRewrite:
    def __init__(self, fromtype, totype):
        self.fromtype = fromtype
        self.totype = totype
        self.uid = None

    def gendecls(self, uid):
        self.uid = str(uid)
        return '''char %s[] = "%s";
char %s[] = "%s";
char %s[] = "%s";
char %s[] = "%s";'''% (
            self.base_name_decl(),
            str(self.fromtype.base_name),
            self.qualifiers_decl(),
            reduce(lambda a, q: q[0] + a, self.fromtype.qualifiers, ''),
            self.to_base_name_decl(),
            str(self.totype.base_name),
            self.to_qualifiers_decl(),
            reduce(lambda a, q: q[0] + a, self.totype.qualifiers, ''),
        )

    def genrule(self):
        return '{ %s, %s, %s, %s }'% (
            self.base_name_decl(), self.qualifiers_decl(), 
            self.to_base_name_decl(), self.to_qualifiers_decl())

    def base_name_decl(self):
        return self.declbase() + '_base_name'

    def qualifiers_decl(self):
        return self.declbase() + '_qualifiers'

    def to_base_name_decl(self):
        return self.declbase() + '_to_base_name'

    def to_qualifiers_decl(self):
        return self.declbase() + '_to_qualifiers'

    def declbase(self):
        return '_type_rule_'+ self.uid

    @classmethod
    def gennullrule(cls):
        return '{ 0, 0, 0, 0 }'

    def __cmp__(self, o):
        '''"<" here means "less specific".'''
        return cmp(self.fromtype, o.fromtype)
    def __hash__(self):
        return hash(repr(self))
    def __repr__(self):
        return '<TypeRewrite from=%r to=%r>'% (self.fromtype, self.totype)


class Pattern:
    def __init__(self, text, argno=None):
        self.text = text
        self.argno = argno

    def iswildcard(self):
        return self.argno is not None

    def __cmp__(self, o):
        '''Return true if |self| < |o|.  The "<" operator here means
        "less specific," so a wildcard pattern is less specific than a
        literal pattern.'''
        if self.iswildcard() == o.iswildcard():
            # alphanumeric sort looks nicer, but doesn't matter
            return cmp(self.text, o.text)
        else:
            return -1 if self.iswildcard() else 1
    def __hash__(self):
        return hash(repr(self))
    def __repr__(self):
        return '%s%r'% ('$'+str(self.argno) if self.iswildcard() else '',
                        self.text)
    def __repr__(self):
        return '%s%s'% ('$'+str(self.argno) if self.iswildcard() else '',
                        self.text)

class CallRewrite:
    def __init__(self, fromcall, tocall):
        self.fromcall = fromcall
        self.tocall = tocall

    def __cmp__(self, o):
        return cmp(self.fromcall, o.fromcall)
    def __hash__(self):
        return hash(repr(self))
    def __repr__(self):
        return '<CallRewrite from=%r to=%r>'% (self.fromcall, self.tocall)

    def gendecls(self, uid):
        self.uid = str(uid)     # uid to use for all this rule's decls
        self._suid = 0          # intra-rule uid for declaring strings

        def strdecl():
            t = self.prog_decl() +'_str%s'% (self._suid)
            self._suid += 1
            return t

        stringdecls = StringIO.StringIO()
        stringdecls.write('const char %s[] = "%s";\n'% (
                self.callable_decl(), self.fromcall.callable))
        progdecl = StringIO.StringIO()
        progdecl.write('WriteInstr %s[] = {\n'% (self.prog_decl()))

        def _emitdeclstr(decl):
            progdecl.write(
                '  { WriteInstr::EMIT_STRING, (uintptr_t) %s },\n'% (decl))
        def emitdot(): _emitdeclstr('dot')
        def emitarrow(): _emitdeclstr('arrow')
        def emitcomma(): _emitdeclstr('comma')
        def emitlparen(): _emitdeclstr('lparen')
        def emitrparen(): _emitdeclstr('rparen')
        def emitnew(): _emitdeclstr('_new')
        def emitdelete(): _emitdeclstr('_delete')

        def emitstring(str):
            sdecl = strdecl()
            stringdecls.write('const char %s[] = "%s";\n'% (sdecl, str))
            _emitdeclstr(sdecl)

        def emitcapture(i):
            progdecl.write(
                '  { WriteInstr::EMIT_CAPT_ARG, (uintptr_t) %s },\n'% (i))

        # generate the tocall callable
        if self.tocall.isnew:
            emitnew()
        elif self.tocall.isdelete:
            emitdelete()

        if isinstance(self.tocall.callable, CxxQualifiedId):
            emitstring(str(self.tocall.callable))
            # TODO support more expr types
        elif isinstance(self.tocall.callable, CxxExprSelect):
            # base obj
            if self.tocall.callable.base.iswildcard():
                emitcapture(self.tocall.callable.base.argno)
            else:
                emitstring(self.tocall.callable.base.text)

            # selector
            if CxxExprSelect.DOT == self.tocall.callable.seltype:
                emitdot()
            else:
                emitarrow()

            # field
            emitstring(self.tocall.callable.field)

        # generate the tocall args, if necessary
        if self.tocall.args is not None:
            # |None| means we replace the function name only
            if not self.tocall.isdelete: emitlparen()
            for i, arg in enumerate(self.tocall.args):
                if arg.iswildcard():
                    emitcapture(arg.argno)
                else:
                    emitstring(arg.text)
                
                if i != (len(self.tocall.args) - 1):
                    emitcomma()
            if not self.tocall.isdelete: emitrparen()

        # done; emit the instruction saying so
        progdecl.write('  { WriteInstr::HALT, (uintptr_t) 0 } };')

        return stringdecls.getvalue() + progdecl.getvalue()

    def genrule(self):
        return '{ %s, %s, %s }'% (self.callable_decl(),
                                  (-1 if self.fromcall.args is None
                                   else len(self.fromcall.args)),
                                  self.prog_decl())

    def callable_decl(self):
        return self.base_decl()+ '_callable'

    def prog_decl(self):
        return self.base_decl()+ '_prog'

    def base_decl(self):
        return '_call_rule_'+ self.uid

    @classmethod
    def gennullrule(cls):
        return '{ 0, 0, 0 }'

    @classmethod
    def patternize(cls, call, wcs=None):
        '''Process |call| into its constituent Patterns, and return them
        in a CxxExprCall.  If there's a token that might be either a 
        wildcard or a pattern, then we check |wcs|.  If it's in there, we
        consider the argument to be a wildcard.
        If |wcs| is None, then we assume ambiguous patterns are wildcards.'''
        pats = { }

        def makepat(*args):
            pat = Pattern(*args)
            if pat.iswildcard():  pats[pat.text] = pat.argno
            return pat

        if isinstance(call, CxxExprCall):
            callable, args = call.callable, call.args 
        elif call.isdelete:
            callable, args = None, call.args

        # patternize callable
        callablepat = None
        if isinstance(callable, CxxQualifiedId):
            # no wildcard: it's specifying a function
            for part in callable:
                if callablepat is None:
                    callablepat = CxxQualifiedId(makepat(part))
                else:  
                    callablepat.qualify(makepat(part))
        elif isinstance(callable, CxxExprSelect):
            # ambiguity: is the base of the selection a wildcard or not?
            # note: this assumes that foo->Bar() can't be a |fromcall|
            # TODO support more complicated expressions
            base = callable.base
            callablepat = CxxExprSelect(
                makepat(base, None if wcs is None else wcs.get(base, None)),
                callable.seltype,
                callable.field)

        # patternize args
        argspats = None
        if args is not None:
            # ambiguity: for each argument, we don't know if it's specifying
            # a wildcard or a name
            argspats = [ ]
            for i, arg in enumerate(args):
                argspats.append(
                    makepat(arg, 
                            i if wcs is None else wcs.get(arg, None)))

        return (CxxExprCall(callablepat, argspats, 
                           ismethodcall=call.ismethodcall, 
                           isnew=call.isnew,
                           isdelete=call.isdelete),
                pats)

    @classmethod
    def create(cls, fromcall, tocall):
        fromcall, fcpats, = cls.patternize(fromcall)
        tocall, _ = cls.patternize(tocall, fcpats)
        return CallRewrite(fromcall, tocall)

##-----------------------------------------------------------------------------

# XXX TESTING
def NYI(what=None):
    what = what or sys._getframe(1).f_code.co_name
    raise NotImplementedError, what

reserved = {
    'call' : 'CALL',
    'class' : 'CLASS',
    'classmethod' : 'CLASSMETHOD',
    'delete' : 'DELETE',
    'function' : 'FUNCTION',
    'global' : 'GLOBAL',
    'member' : 'MEMBER',
    'method' : 'METHOD',
    'new' : 'NEW',
    'rewrite' : 'REWRITE',
    'type' : 'TYPE',
}
tokens = [
    'COLONCOLON', 'EQARROW', 'DASHARROW',
    'COMMA', 'DOT', 'STAR', 'AMPERSAND', 'NL',
    'LBRACKET', 'RBRACKET', 'LBRACE', 'RBRACE', 'LPAREN', 'RPAREN', 
    'ID'
] + list(reserved.values())

t_COLONCOLON = r'\:\:'
t_EQARROW = r'\=>'
t_DASHARROW = r'\->'
t_COMMA = r'\,'
t_DOT = r'\.'
t_STAR = r'\*'
t_AMPERSAND = r'\&'
t_LBRACKET = r'\['
t_RBRACKET = r'\]'
t_LBRACE = r'\{'
t_RBRACE = r'\}'
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_NL = r'(?:\r\n|\r|\n)'

t_ignore = ' \f\t\v'

def t_ID(t):
    r'[a-zA-Z_][a-zA-Z0-9_]*'
    t.type = reserved.get(t.value, 'ID')
    return t

def t_error(t):
    raise Exception, 'bad character '+ str(t)

import ply.lex as lex
lex.lex(lextab="porky_lextab")

def p_RewriteSpec(p):
    '''RewriteSpec : REWRITE MaybeNewlines ID MaybeNewlines LBRACE MaybeNewlines RewriteList RBRACE MaybeNewlines'''
    p[0] = Rewriter.create(p[3], p[7])

def p_RewriteList(p):
    '''RewriteList : RewriteList OneRewrite
                   | OneRewrite'''
    if 2 == len(p):
        p[0] = [ p[1] ]
    else:
        p[1].append(p[2])
        p[0] = p[1]
def p_OneRewrite(p):
    '''OneRewrite : ClassRewrite Newlines
                  | FunctionRewrite Newlines
                  | CallRewrite Newlines
                  | TypeRewrite Newlines
                  | GlobalVarRewrite Newlines'''
    p[0] = p[1]

def p_ClassRewrite(p):
    '''ClassRewrite : CLASS ID EQARROW ID
                    | CLASS ID ClassBodyRewrite
                    | CLASS ID EQARROW ID ClassBodyRewrite'''
    NYI()

def p_ClassBodyRewrite(p):
    '''ClassBodyRewrite : LBRACE ClassBodyRewriteList RBRACE'''
    NYI()

def p_ClassBodyRewriteList(p):
    '''ClassBodyRewriteList : ClassBodyRewriteList OneClassBodyRewrite Newlines
                            | OneClassBodyRewrite Newlines'''
    NYI()
def p_OneClassBodyRewrite(p):
    '''OneClassBodyRewrite : METHOD ID EQARROW ID
                           | METHOD ID Callargs EQARROW ExprCall
                           | CLASSMETHOD ID EQARROW ID
                           | CLASSMETHOD ID Callargs EQARROW ExprCall
                           | MEMBER ID EQARROW ID'''
    NYI()

def p_FunctionRewrite(p):
    '''FunctionRewrite : FUNCTION QualifiedId EQARROW QualifiedId'''
    NYI()

def p_CallRewrite(p):
    '''CallRewrite : CALL QualifiedId EQARROW QualifiedId
                   | CALL ExprFunCall EQARROW ExprCall'''
    p[0] = CallRewrite.create(p[2], p[4])

def p_ExprCall(p):
    '''ExprCall : ExprFunCall
                | ExprMethodCall
                | ExprNew
                | ExprDelete'''
    p[0] = p[1]

def p_ExprFunCall(p):
    '''ExprFunCall : QualifiedId Callargs'''
    p[0] = CxxExprCall(p[1], p[2])

def p_ExprMethodCall(p):
    '''ExprMethodCall : ExprSelect Callargs'''
    p[0] = CxxExprCall(p[1], p[2], ismethodcall=True)

# TODO ExprNew and ExprDelete assume that they are only used as 'tocall's,
# and not as patterns.  this a rather hacky way of doing things
def p_ExprNew(p):
    '''ExprNew : NEW QualifiedId Callargs'''
    p[0] = CxxExprCall(p[2], p[3], isnew=True);

def p_ExprDelete(p):
    '''ExprDelete : DELETE ID'''
    p[0] = CxxExprCall(None, [ p[2] ], isdelete=True);

def p_Callargs(p): 
    '''Callargs : LPAREN RPAREN
                | LPAREN IdList RPAREN
                '''
    if 3 == len(p):
        p[0] = []
    else:
        p[0] = p[2]

# TODO recursive select             
def p_ExprSelect(p):
    '''ExprSelect : ID Selector ID'''
    p[0] = CxxExprSelect(p[1], p[2], p[3])

def p_Selector(p):
    '''Selector : DOT 
                | DASHARROW'''
    p[0] = CxxExprSelect.DOT if '.' == p[1] else CxxExprSelect.PDOT

def p_TypeRewrite(p):
    '''TypeRewrite : TYPE Type EQARROW Type'''
    p[0] = TypeRewrite(p[2], p[4])

def p_GlobalVarRewrite(p):
    '''GlobalVarRewrite : GLOBAL QualifiedId EQARROW QualifiedId'''
    NYI()

def p_Type(p):
    '''Type : QualifiedId
            | Type STAR
            | Type AMPERSAND
            | Type LBRACKET RBRACKET'''
    # TODO fix ^^^; can't have T&&, e.g.
    # TODO support templates; maybe with wildcard instantiations <*, *>
    if 2 == len(p):
        p[0] = CxxType(p[1])
    elif 3 == len(p):
        if '*' is p[2]:
            p[1].pointer()
        elif '&' is p[2]:
            p[1].reference()
        p[0] = p[1]
    else:
        # TODO handling these in source translation is unnecessarily tricky.
        # thanks, Kernighan and Ritchie.
        NYI('array types')

def p_IdList(p):
    '''IdList : IdList COMMA ID
               | ID'''
    if 2 == len(p):
        p[0] = [ p[1] ]
    else:
        p[1].append(p[3])
        p[0] = p[1]

def p_QualifiedId(p):
    '''QualifiedId : ID
                   | COLONCOLON ID
                   | QualifiedId COLONCOLON ID'''
    # TODO support templates
    if 2 == len(p):
        p[0] = CxxQualifiedId(p[1])
    elif 3 == len(p):
        # TODO is this correct? '::Foo' means "Foo in global namespace"?
        p[0] = CxxQualifiedId(p[2])
    else:
        p[1].qualify(p[3])
        p[0] = p[1]

def p_MaybeNewlines(p):
    '''MaybeNewlines : Newlines
                     | '''      # empty
    return

def p_Newlines(p):
    '''Newlines : Newlines NL
                | NL'''
    return

def p_error(t):
    raise Exception, 'syntax error near "%s"'% (t.value)

import ply.yacc as yacc
yacc.yacc(tabmodule="porky_yacctab", debugfile="porky_yacctab_lrtable")

##-----------------------------------------------------------------------------
import fnmatch, getopt, os, sys, tempfile
from commands import getstatusoutput as _

try:
    import pork_barrel
except Exception:
    print >>sys.stderr, '''Can't locate `pork_barrel.py'.
Make sure it can be found in your $PYTHONPATH.'''
    sys.exit(1)
    
def usage(err):
    out, rv = (sys.stderr, 1) if err else (sys.stdout, 0)
    print >>out, '''
Usage:  porky [OPTIONS...] [REWRITE_SPEC] [FILES_TO_REWRITE...]

where OPTIONS are zero or more of
  -j N    : use N-way parallelism while rewriting.
            default: -j 1
  -h, --help
          : print this message and exit
  -o PATH, --output=PATH
          : where to save the generated patch.  Must be writeable.  `-' means
              stdout.
            default: stdout
  -p STYLE, --ptr-style=STYLE
          : style in which to emit rewritten pointer types.  One of:
                -p `*_',  for `Foo* f'
                -p `_*',  for 'Foo *f'
            default: -p `*_'
  -r STYLE, --ref-style=STYLE
          : style in which to emit rewritten reference types.  One of:
                -r `&_',  for `Foo& f'
                -r `_&',  for 'Foo &f'
            default: -r `&_'
  -S      : just generate and save the pork C++ code; don't run the rewriter.
  -v N    : be as verbose as N.  set to 0 for no output, and higher than 1
              for more output.
            default: -v 1

and REWRITE_SPEC is NONE or ONE of
  -e EXPR, --expr=EXPR
          : generate a temporary rewriter from the rewrite expression EXPR
  -m PROGRAM, --program=PROGRAM
          : generate a named rewriter from the rewrite program PROGRAM.
            PROGRAM can be `-', in which case the program is read from stdin.
  default : read specification from stdin

[TODO brief example of rewrite grammar?]

and FILES_TO_REWRITE is
  SHELL_PATTERN
          : e.g., `foo/bar.cc', `~/code/*', '$HOME/code'
              CAREFUL: these are expanded relative to the current directory.
              It's best to specify an absolute pattern.
  -x SHELL_PATTERN, --exclude=SHELL_PATTERN
          : don't attempt to rewrite any programs matching `SHELL_PATTERN'
'''
    sys.exit(rv)

if __name__ == '__main__':
    # TODO ugh, requiring a directory to run from kinda sucks
    pigletdir = os.path.abspath(os.getenv("PIGLET_DIR", "../piglet"))
    config_mk = os.path.join(pigletdir, 'config.mk')
    build_mk = os.path.join(pigletdir, 'build.mk')
    if not (os.path.isdir(pigletdir)
            and os.path.isfile(config_mk)
            and os.path.isfile(build_mk)):
        print >>sys.stderr, '''
Can't locate the `$PORK/piglet/' directory, or can't find `build.mk' there.
Please 
  (1) either set the environment variable `PIGLET_DIR' or run %s from
      within `$PORK/porky/`
  (2) ensure that pork and piglet have been installed and configured
'''
        sys.exit(1)

    try:
        opts, args = getopt.gnu_getopt(
            sys.argv[1:],
            'j:ho:p:r:Sv:e:m:x:',
            ['--output=', '--ptr-style=', '--ref-style=', 
             '--expr=', '--program=',
             '--exclude='])
    except getopt.GetoptError, err:
        print >>sys.stderr, str(err), '\n'
        usage(err=True)

    ncpus = 1
    patchout = '-'
    ptrstyle = '* '
    refstyle = '* '
    genonly = False
    savetempsto = None
    expr = None
    prog = None
    exfilepats = [ ]
    filepats = args

    for o, a in opts:
        if o in ('-j',):
            ncpus = int(a)
        elif o in ('-h', '--help'):
            usage(err=False)
        elif o in ('-o', '--output'):
            if a != '-' and not os.access(patchout, os.W_OK):
                print >>sys.stderr, "don't have write access to patch file", a
                usage(err=True)
            patchout = a
        elif o in ('-p', '--ptr-style'):
            if a not in ('*_', '_*'):
                print >>sys.stderr, 'invalid pointer style', repr(a)
                usage(err=True)
            ptrstyle = a
        elif o in ('-r', '--ref-style'):
            if a not in ('&_', '_&'):
                print >>sys.stderr, 'invalid reference style', repr(a)
                usage(err=True)
            refstyle = a
        elif o in ('-S',):
            genonly = True
        elif o in ('-v',):
            _verbosity = int(a)

        elif o in ('-e', '--expr'):
            if expr or prog:
                print >>sys.stderr, 'second rewrite spec seen', repr(a)
                usage(err=True)
            expr = o
        elif o in ('-m', '--program'):
            if expr or prog:
                print >>sys.stderr, 'second rewrite spec seen', repr(a)
                usage(err=True)
            if not os.access(a, os.R_OK):
                print >>sys.stderr, "don't have read access to program", a
                usage(err=True)
            prog = a

        elif o in ('-x', '--exclude'):
            exfilepats.append(a)

    specstring = None
    if expr is not None:
        log (1, 'using literal rewrite expression')
        specstring = '''rewriter {
%s
}'''% ('\n'.join(expr.split(';')))
    elif prog is not None:
        log (1, 'using rewrite program %s', prog)
        specstring = open(prog, 'r').read()
    else:
        log(1, 'reading rewrite spec from stdin')
        specstring = sys.stdin.read()

    log(3, 'rewrite spec:\n%s', specstring)


    ##--------------------------------------------------
    log(1, 'parsing input specification')
    ##--------------------------------------------------    
    rwspec = yacc.parse(input=specstring, debug=1)


    ##--------------------------------------------------
    log(1, 'generating C++ pork rewriter')
    ##--------------------------------------------------
    codedir = tempfile.mkdtemp('.code', rwspec.name)
    rewriterCxxFilename = os.path.join(codedir, rwspec.name + '.cc')

    log(2, '  writing code to %s', rewriterCxxFilename)

    rewriterCxxFile = open(rewriterCxxFilename, 'w')
    rwspec.gencode(out=rewriterCxxFile)
    rewriterCxxFile.close()    

    makefile = os.path.join(codedir, 'Makefile')
    open(makefile, 'w').write(
        MAKEFILE_TEMPLATE.substitute({
                'pigletdir_config_mk': config_mk,
                'pigletdir_build_mk': build_mk,
                'rewriter': rwspec.name,
                }))

    if genonly:
        print 'find generated code at %s'% (codedir)
        sys.exit(0)

    ##--------------------------------------------------
    log(1, 'compiling generated code')
    ##--------------------------------------------------
    cmd = "make -C '%s'"% (codedir)

    log (2, '  invoking make as "%s"'% (cmd))

    status, makeoutput = _(cmd)
    if os.WIFEXITED(status) and os.WEXITSTATUS(status):
        print >>sys.stderr, 'make exited abnormally; output:'
        sys.stderr.write(makeoutput)
        sys.exit(1)
    elif os.WIFSIGNALED(status):
        print >>sys.stderr, 'make exited from signal; output:'
        sys.stderr.write(makeoutput)
        sys.exit(1)


    ##--------------------------------------------------
    log(1, 'generating list of files to process')
    ##--------------------------------------------------
    if len(filepats) == 0:
        print >>sys.stderr, 'need some files to rewrite'
        usage(err=True)

    def normalizepath(p):
        return os.path.abspath(os.path.expandvars(os.path.expanduser(os.path.normpath(p))))

    files = set()
    for fpat in [ normalizepath(f) for f in filepats ]:
        dir = os.path.dirname(fpat)

        log(2, '  walking %s, trying to match %s', dir, fpat)

        def addfiles(a, d, names):
            for name in names:
                fullpath = os.path.join(d, name)
                if fnmatch.fnmatch(fullpath, fpat):
                    files.add(fullpath)
        os.path.walk(dir, addfiles, None)

    log(2, '  matched %s files', len(files))

    # Nb: we can't filter input files here, because we have no idea what
    # they may #include.  some of these headers we indeed may want to patch.

    ##--------------------------------------------------
    log(1, 'invoking pork barrel to manage rewrite')
    ##--------------------------------------------------
    cmd = "'%s' '%%(file)s'"% (os.path.join(codedir, rwspec.name))

    if '-' == patchout:
        log(1, 'patch will be written to stdout')
        patchout = sys.stdout
    else:
        log(1, 'patch will be written to %s', patchout)
        patchout = open(patchout, 'w')

    pork_barrel.run(
        files, 
        cmd,
        exfilepats,
        None,                   # we've already decided which files to include
        patchout, 
        ncpus,
        sys.stderr,
        _verbosity)

    # blow away temp rewriter, if desired
    cmd = "rm -rf '%s'"% (codedir)
    print '\nAbout to execute "%s" to remove temporary files.  Is this OK?'% (cmd)

    if 'yes' == raw_input('Please enter "yes" or "no" > '):
        status, rmoutput = _(cmd)
        log(2, '  cleaned up: %s', rmoutput)
    else:
        print 'Not deleting temporaries'

    log(1, "That's all, folks!")
